%
% Broyden method for non linear system with damping
%
function varargout = BroydenDamped( F, x0, epsi, varargin )
  global BroydenDamped_global ;
  BroydenDamped_global . F = F ;

  % if required save the computed iteration
  if nargout == 2
    varargout{1} = [] ;
    varargout{2} = [] ;
  end ;

  linesearch = 'wolfe' ;
  if nargin >= 5 ; linesearch = varargin{1} ; end ;

  % initialize Broyden Matrix
  x  = x0 ;
  Fx = feval( F, x ) ;
  A  = diag(1E-6+abs(Fx)) ;

  for k=1:1000

    % if required save the computed iteration
    if nargout == 2
      varargout{1} = [varargout{1}, norm(feval( F, x ),inf)]  ;
      varargout{2} = [varargout{2}, x ] ;
    end ;

    % check for bad termination
    rc = rcond(A) ;
    if ~ isfinite(rc) | rc < 1E-10
      disp('BroydenDamped: bad termination');
      return ;
    end ;

    % compute search direction
    d = -A\Fx ;

    % Setup initial point and direction for linesearch
    BroydenDamped_global . x0 = x ;
    BroydenDamped_global . d  = d ;
    BroydenDamped_global . A  = A ;
    
    % perform selected linesearch
    switch lower(linesearch)
      case 'wolfe'  ; alpha = WolfeLineSearch( @Fdir, @DFdir ) ;
      case 'armijo' ; alpha = ArmijoLineSearch( @Fdir, @DFdir ) ;
      case 'simple' ; alpha = ArmijoSimpleSearch( @Fdir, @DFdir ) ;
      otherwise ; error( ['NonlinearCG bad linesearch: ' linesearch] ) ;
    end

    disp([k alpha])

    % update x
    s    = alpha * d ;
    x    = x + s ;
    
    % update matrix    
    Fnew = feval( F,  x ) ;
    y    = Fnew - Fx ;
    Fx   = Fnew ;

    A = A + (y-A*s) * (s' / dot(s,s)) ;    
    
    % check for termination
    if norm( d, inf ) < epsi & norm(Fx) < epsi
      break ;
    end ;

  end ;

  if nargout == 1
    varargout{1} = x ;
  end ;

end

function res = Fdir( t )
  global BroydenDamped_global ;
  F    = BroydenDamped_global . F ;
  A    = BroydenDamped_global . A ;
  x0   = BroydenDamped_global . x0 ;
  d    = BroydenDamped_global . d ;
  FF   = A\feval(F, x0 + t * d) ;
  res  = norm(FF,inf) ;
end

function res = DFdir( t )
  h    = 1E-4 ;
  res  = (Fdir(t+h)-Fdir(t-h))/(2*h) ;
end
