%
% Perform Wolfe Line Search
%
function l = WolfeLineSearch( fun, Dfun, varargin )

  lMin    = 1E-8 ;
  lMax    = 1E+8 ;
  epsi    = 1E-4 ;
  c1      = 1E-4 ;
  c2      = 0.5  ;
  dumpMin = 2    ;
  dumpMax = 100  ;

  if nargin >= 3 
    f0 = varargin{1} ;
  else
    f0 = feval( fun, 0 ) ;
  end

  Df0 = feval( Dfun, 0 ) ;
  l   = 1 ;
  
  while l > lMin
    fl = feval( fun, l ) ;
    if fl <= f0 + (l*c1) * Df0;
      % goto ZOOM
      break ;
    else
      if l == 1
        lTmp = quadratic( f0, Df0, fl, l ) ;
      else
        lTmp = cubic( f0, Df0, fl, l, fp, p ) ;
      end ;
      p  = l ;
      fp = fl ;
      l  = max( min( lTmp, l/dumpMin ), l/dumpMax ) ;
    end ;
  end ;

  if l < lMin
    disp( 'failed search 0' ) ;
    l = lMin ; % Failed search
    return ;
  end ;

  %   ________   ___  __  __ 
  %  |__  / _ \ / _ \|  \/  |
  %    / / | | | | | | |\/| |
  %   / /| |_| | |_| | |  | |
  %  /____\___/ \___/|_|  |_|
  %
  Dfl = feval( Dfun, l ) ;
  if Dfl >= c2 * Df0 ;
    return ; % found Wolfe point
  end ;

  if l == 1
    % forward search
    while l <= lMax 
      p  = l ;
      fp = fl ;
      l  = 2*l ;
      fl = feval( fun, l ) ;
      if fl > f0 + (c1*l) * Df0
        p  = l ;
        fp = fl ;
        % goto REFINE
        break ;
      end ;
      Dfl = feval( Dfun, l ) ;
      if Dfl >= c2 * Df0 ;
        return ; % found Wolfe point
      end ;
    end ;    
  end ;

  if l > lMax
    disp( 'failed search 1' ) ;
    l = lMax ; % failed search
    return ;
  end;

  %   ____  _____ _____ ___ _   _ _____ 
  %  |  _ \| ____|  ___|_ _| \ | | ____|
  %  | |_) |  _| | |_   | ||  \| |  _|  
  %  |  _ <| |___|  _|  | || |\  | |___ 
  %  |_| \_\_____|_|   |___|_| \_|_____|
  %
  lLo  = l ;
  fLo  = fl ;
  DfLo = Dfl ;

  lHi  = p ;
  fHi  = fp ;

  Delta = lHi - lLo ;
  
  while abs(Delta) > epsi

    deltaLambda = quadratic( fLo, DfLo, fHi, Delta ) ;
    if deltaLambda > 0
      deltaLambda = min( Delta/dumpMin, max( Delta/dumpMax, deltaLambda )) ;
    else
      deltaLambda = max( Delta/dumpMin, min( Delta/dumpMax, deltaLambda )) ;
    end
    l  = lLo + deltaLambda ;
    fl = feval( fun, l ) ;

    if fl > f0 + (l*c1) * Df0 | fl > fLo
      lHi   = l ;
      fHi   = fl ;
    else
      Dfl = feval( Dfun, l ) ;

      if Dfl >= c2 * Df0;
        return ; % found Wolfe point
      end ;

      lLo  = l ;
      fLo  = fl ;
      DfLo = Dfl ;
    end
    Delta = lHi - lLo ;
  end
  disp( 'failed search 2' ) ;
  return ; % failed search
end
%                         _           _   _      
%    __ _ _   _  __ _  __| |_ __ __ _| |_(_) ___ 
%   / _` | | | |/ _` |/ _` | '__/ _` | __| |/ __|
%  | (_| | |_| | (_| | (_| | | | (_| | |_| | (__ 
%   \__, |\__,_|\__,_|\__,_|_|  \__,_|\__|_|\___|
%      |_|                                       
%      
function alpha = quadratic( f0, Df0, fp, p )
  alpha = Df0 * p^2 / ( 2*(f0+Df0*p-fp) ) ;
end
%              _     _      
%    ___ _   _| |__ (_) ___ 
%   / __| | | | '_ \| |/ __|
%  | (__| |_| | |_) | | (__ 
%   \___|\__,_|_.__/|_|\___|
% 
function alpha = cubic( f0, Df0, fl, l, fp, p )
  % f0 + Df0 * x + a * x^2 + b * x^3

  bf1 = fl - (f0+l*Df0) ;
  bf2 = fp - (f0+p*Df0) ;
  bf3 = (p*l)^2*(p-l) ;
  a   = (p^3 * bf1 - l^3 * bf2) / bf3 ;
  b   = (l^2 * bf2 - p^2 * bf1) / bf3 ;

  if abs(b) < 1E-10
    % points disposed near a parabola
    alpha = -Df0 / (2*a) ;
  else
    delta = a^2 - 3*b*Df0 ;
    if delta < 0
      % complex root, using quadratic again
      alpha = quadratic( f0, Df0, fl, l ) ;
    else
      % regular case, using cubic
      alpha = (sqrt(delta)-a)/(3*b) ;
    end;
  end;
end

